
-- Block ID -> Geometry description for Eihort 0.3.0+

-- The first part of this file contains helper functions used to construct
-- the various block geometry objects.
-- The actual Block ID -> Geometry list is at the end of the file

require "texpack";

function loadBlockDesc()
	local blocks = eihort.newBlockDesc();

	-- Basic texture loading
	local textureCache = { };
	local defAniso = 'aniso_' .. tostring( Config.anisotropy or 0 );
	local function texUpload( img, ... )
		return img:uploadToGL( 'repeat', 'mag_nearest', 'min_linear', 'mip_linear', 'mipgen_alphawt_box', defAniso, ... );
	end
	function TEX( filename, ... )
		local cached = textureCache[filename];
		if cached then
			return cached;
		end
		local path, file = string.match( filename, "^(.-)([^/]*)$" );
		local teximg = loadTexture( path, file, true );
		local tex = texUpload( teximg, ... );
		textureCache[filename] = tex;
		teximg:destroy();
		return tex;
	end
	function BTEX( blocktex, ... )
		return TEX( "textures/blocks/" .. blocktex .. ".png", ... );
	end

	-- Some special cases where the alpha channels must be modified
	local function Tex_AlphaFromGray( t1, t2, ... )
		t1:grayToAlpha( t2 );
		local tex = texUpload( t1, 'mipgen_box', ... );
		t1:destroy();
		t2:destroy();
		return tex;
	end
	local function BTEX_AlphaFromGray( basetex, graytex, ... )
		return Tex_AlphaFromGray( loadTexture( "textures/blocks/", basetex .. ".png", true ), loadTexture( "textures/blocks/", graytex .. ".png", true ), ... );
	end
	local function BTEX_InAlpha( alphatex, ... )
		local img = loadTexture( "textures/blocks/", alphatex .. ".png", true );
		return Tex_AlphaFromGray( eihort.newImage( img.width, img.height, 0, 0, 0, 0 ), img, ... );
	end
	local function BTEX_NoAlpha( blocktex, ... )
		local img = loadTexture( "textures/blocks/", blocktex .. ".png", true );
		return Tex_AlphaFromGray( img, eihort.newImage( img.width, img.height, 0, 0, 0, 0 ), ... );
	end

	-- Special cases for water, lava and portals
	local function BTEX_Water( watertex, columns, ... )
		columns = columns or 1;
		local teximg = loadTexture( "textures/blocks/", watertex .. ".png", true );
		local tex = texUpload( teximg:sub( 0, 0, teximg.width / columns, teximg.width / columns ), ... );
		teximg:destroy();
		return tex;
	end

	-- Load chest.png
	local chestImage = loadTexture( "item/", "chest.png", true );
	if math.fmod( chestImage.width, 64 ) ~= 0 or math.fmod( chestImage.height, 64 ) ~= 0 then
		error( "chest.png's dimensions must be multiples of 64." );
	end
	function ChestPNG( x, y, ... )
		local tw = chestImage.width/4;
		local th = chestImage.height/4;
		local tex = eihort.newImage( tw, th, 0, 0, 0, 0 );
		assert(y<4);
		if y == 0 then
			local t1 = chestImage:sub( tw*x*14/16, 0, tw*14/16, th*14/16 );
			tex:put( t1, tw*1/16, th*1/16 );
			t1:destroy();
		elseif y == 1 then
			local t1 = chestImage:sub( x*tw*14/16, th*14/16, tw*14/16, th*5/16 );
			local t2 = chestImage:sub( x*tw*14/16, th*34/16, tw*14/16, th*9/16 );
			tex:put( t1, tw*1/16, th*2/16 );
			tex:put( t2, tw*1/16, th*7/16 );
			t1:destroy();
			t2:destroy();
		else
			local t1 = chestImage:sub( tw*x*14/16, th*19/16, tw*14/16, th*14/16 );
			tex:put( t1, tw*1/16, th*1/16 );
			t1:destroy();
		end
		local tx = texUpload( tex, ... );
		tex:destroy();
		return tx;
	end
	chestSide = ChestPNG( 0, 1 );
	chestTop = ChestPNG( 1, 0 );


	-- Block geometry adapters
	local function DataAdapter( mask, ... ) -- "mask" is the bitmask for the data value.
		local geoms = { ... };
		local geomList = { };
		local solidity = 0x3f;
		for i, v in ipairs( geoms ) do
			if type( v ) == "number" then
				geomList[i] = geomList[v+1];
			else
				geomList[i] = v[1];
				solidity = eihort.intAnd( solidity, v[2] );
			end
		end
		return { eihort.geom.dataAdapter( mask, unpack( geomList ) ), solidity };
	end
	local function RotatingAdapter( normalGeom, faceGeom )
		return { eihort.geom.rotatingAdapter( normalGeom[1], faceGeom[1] ),
			     eihort.intAnd( normalGeom[2], eihort.intOr( 0x60, faceGeom[2] ) ) };
	end
	local function FaceBitAdapter( faceGeom )
		return { eihort.geom.faceBitAdapter( faceGeom[1] ), 0x00 };
	end
	local function FacingAdapter( normalGeom, faceGeom )
		if not normalGeom then
			normalGeom = { false, 0x00 };
		end
		return { eihort.geom.facingAdapter( normalGeom[1], faceGeom[1] ),
			     eihort.intAnd( normalGeom[2], eihort.intOr( 0x60, faceGeom[2] ) ) };
	end
	local function TopDifferentAdapter( normalGeom, diffGeom, topId, topGeom )
		return { eihort.geom.topDifferentAdapter( normalGeom[1], diffGeom[1], topId, topGeom and topGeom[1] ),
		         eihort.intAnd( normalGeom[2], diffGeom[2] ) };
	end
	local function SetTexScale( geom, s, t )
		geom[1]:setTexScale( s, t );
		return geom;
	end
	local function DelayRender( geom, delta )
		geom[1]:renderGroupAdd( delta or 1 );
		return geom;
	end

	-- Block Geometry creators
	-- In all the following, ... represents a list of textures with different
	-- meaning depending on how many textures are given:
	--     One texture: All sides of the block are covered with the same texture
	--     Two textures: The four sides take the first texture, the top and
	--                   bottom take the second
	--     Three textures: Sides, bottom, top
	--     Four textures: First two textures are the sides, which go on alternate
	--                    faces, last two textures are bottom and top
	--     Six textures: All 6 sides in the order X- X+ Z- Z+ Y- Y+
	local function OpaqueBlock( ... )
		return { eihort.geom.opaqueBlock( ... ), 0x3f };
	end
	local function HollowOpaqueBlock( ... )
		return { eihort.geom.opaqueBlock( ... ), 0x00 };
	end
	local function BrightOpaqueBlock( ... )
		return { eihort.geom.brightOpaqueBlock( ... ), 0x3f };
	end
	local function TransparentBlock( order, ... )
		return { eihort.geom.transparentBlock( order, ... ), 0x00 };
	end
	local function Slab( topOffset, bottomOffset, ... )
		local solidity = 0;
		if topOffset == 0 then
			solidity = solidity + 0x10;
		end
		if bottomOffset == 0 then
			solidity = solidity + 0x20;
		end
		return { eihort.geom.squashedBlock( topOffset, bottomOffset, ... ), solidity };
	end
	local function CompactedBlock( offsetXn, offsetXp, offsetZn, offsetZp, offsetYn, offsetYp, ... )
		return { eihort.geom.compactedBlock( offsetXn, offsetXp, offsetZn, offsetZp, offsetYn, offsetYp, ... ), 0x00 };
	end
	local function MultiBlockInBlock( offsets, ... )
		return { eihort.geom.multiBlockInBlock( offsets, ... ) , 0x00 };
	end
	local function MultiCompactedBlock( x, y, z, ... )
		return { eihort.geom.multiCompactedBlock( x, y, z, ... ) , 0x00 };
	end
	local function BiomeOpaqueBlock( biomeChannel, ... )
		return { eihort.geom.biomeOpaqueBlock( biomeChannel, ... ), 0x3f };
	end
	local function BiomeHollowOpaqueBlock( biomeChannel, ... )
		return { eihort.geom.biomeOpaqueBlock( biomeChannel, ... ), 0x00 };
	end
	local function BiomeAlphaOpaqueBlock( biomeChannel, ... )
		return { eihort.geom.biomeAlphaOpaqueBlock( biomeChannel, ... ), 0x3f };
	end
	local function Portal( tex )
		return { eihort.geom.portal( tex ), 0x00 };
	end
	local function Cactus( offset, ... )
		return { eihort.geom.cactus( offset, ... ), 0x00 };
	end
	local function BiomeCactus( biomeChannel, offset, ... )
		return { eihort.geom.biomeCactus( biomeChannel, offset, ... ), 0x00 };
	end
	local function Rail( straight, turn )
		return { eihort.geom.rail( straight, turn ), 0x00 };
	end
	local function Door( bottom, top )
		return DataAdapter( 0x8,
			{ eihort.geom.door( bottom ), 0x00 },
			{ eihort.geom.door( top ), 0x00 } );
	end
	local function Stairs( ... )
		return { eihort.geom.stairs( ... ), 0x00 };
	end
	local function Torch( tex )
		return { eihort.geom.torch( tex ), 0x00 };
	end
	local function Flower( tex )
		return { eihort.geom.flower( tex ), 0x00 };
	end
	local function BiomeFlower( biomeChannel, tex )
		return { eihort.geom.biomeFlower( biomeChannel, tex ), 0x00 };
	end
	local function Fence( tex )
		return MultiCompactedBlock(
		          {  10,  10,  -7,  -7, -11,  -2,
				     10,  10,  -7,  -7,  -5,  -8 },
				  {  -7,  -7,  10,  10, -11,  -2,
				     -7,  -7,  10,  10,  -5,  -8 },
				  {  -6,  -6,  -6,  -6,  -0,  -0 },
				  tex );
	end

	-------------------------------------------------------------------
	-- The actual block geometry list
	FurnaceBody = OpaqueBlock( BTEX("furnace_side"), BTEX("furnace_top") );
	MinecraftBlocks = {
		-- [<block id>] = <geometry>;
		[1]   = OpaqueBlock( BTEX("stone") ); -- Stone
		[2]   = TopDifferentAdapter( -- Grass
		          BiomeAlphaOpaqueBlock( 0, BTEX_AlphaFromGray( "grass_side", "grass_side_overlay" ), BTEX_NoAlpha("dirt"), BTEX_InAlpha("grass_top") ), -- Normal grass
		          OpaqueBlock( BTEX("snow_side"), BTEX("dirt") ), 78 ); -- Snowy grass
		[3]   = OpaqueBlock( BTEX("dirt") ); -- Dirt
		[4]   = OpaqueBlock( BTEX("stonebrick") ); -- Cobblestone
		[5]   = DataAdapter( 0x3, -- Wooden Plank
		          OpaqueBlock( BTEX("wood") ), -- oak
		          OpaqueBlock( BTEX("wood_spruce") ), -- spruce
		          OpaqueBlock( BTEX("wood_birch") ), -- birch
		          OpaqueBlock( BTEX("wood_jungle") ) ); -- jungle
		[6]   = DataAdapter( 0x3, -- Sapling
		          Flower( BTEX("sapling") ), -- oak
		          Flower( BTEX("sapling_spruce") ), -- spruce
		          Flower( BTEX("sapling_birch") ), -- birch
		          Flower( BTEX("sapling_jungle") ) ); -- jungle
		[7]   = OpaqueBlock( BTEX("bedrock") ); -- Bedrock
		[8]   = TransparentBlock( 1, BTEX_Water( "water_flow" ), BTEX_Water( "water" ) ); -- Water
		[9]   = TransparentBlock( 2, BTEX_Water( "water" ) ); -- Stationary Water
		[10]  = BrightOpaqueBlock( BTEX_Water( "lava_flow", 2 ), BTEX_Water( "lava" ) ); -- Lava
		[11]  = BrightOpaqueBlock( BTEX_Water( "lava" ) ); -- Stationary Lava
		[12]  = OpaqueBlock( BTEX("sand") ); -- Sand
		[13]  = OpaqueBlock( BTEX("gravel") ); -- Gravel
		[14]  = OpaqueBlock( BTEX("oreGold") ); -- Gold Ore
		[15]  = OpaqueBlock( BTEX("oreIron") ); -- Iron Ore
		[16]  = OpaqueBlock( BTEX("oreCoal") ); -- Coal Ore
		[17]  = DataAdapter( 0x3, -- Wood
		          OpaqueBlock( BTEX("tree_side"), BTEX("tree_top") ), -- oak
		          OpaqueBlock( BTEX("tree_spruce"), BTEX("tree_top") ), -- spruce
		          OpaqueBlock( BTEX("tree_birch"), BTEX("tree_top") ), -- birch
		          OpaqueBlock( BTEX("tree_jungle"), BTEX("tree_top") ) ); -- jungle
		[18]  = DataAdapter( 0x3, -- Leaves
		          BiomeHollowOpaqueBlock( 2, BTEX("leaves") ), -- oak
		          BiomeHollowOpaqueBlock( 2, BTEX("leaves_spruce") ), -- spruce
		          BiomeHollowOpaqueBlock( 1, BTEX("leaves") ), -- birch
		          BiomeHollowOpaqueBlock( 1, BTEX("leaves_jungle") ) ); -- jungle
		[19]  = OpaqueBlock( BTEX("sponge") ); -- Sponge
		[20]  = HollowOpaqueBlock( BTEX("glass") ); -- Glass
		[21]  = OpaqueBlock( BTEX("oreLapis") ); -- Lapis Lazuli Ore
		[22]  = OpaqueBlock( BTEX("blockLapis") ); -- Lapis Lazuli Block
		[23]  = FacingAdapter( FurnaceBody, OpaqueBlock( BTEX("dispenser_front") ) ); -- Dispenser
		[24]  = DataAdapter( 0x3, -- Sandstone
		          OpaqueBlock( BTEX("sandstone_side"), BTEX("sandstone_bottom"), BTEX("sandstone_top") ), -- normal
		          OpaqueBlock( BTEX("sandstone_carved"), BTEX("sandstone_top"), BTEX("sandstone_top") ), -- chiseled
		          OpaqueBlock( BTEX("sandstone_smooth"), BTEX("sandstone_top"), BTEX("sandstone_top") ), -- smooth
		          0 );
		[25]  = OpaqueBlock( BTEX("musicBlock") ); -- Note block
		-- [26] -- Bed
		[27]  = DataAdapter( 0x8, -- Powered rails
		          Rail( BTEX("activatorRail") ), -- not powered
		          Rail( BTEX("activatorRail_powered") ) ); -- powered
		[28]  = Rail( BTEX("detectorRail") ); -- Detector rails
		-- Sticky Piston needs special code: the trick is that the side tiles need to rotate to face the front piece TODO
		[29]  = OpaqueBlock( BTEX("piston_side"), BTEX("piston_bottom"), BTEX("piston_top_sticky") ); -- Sticky Piston up
		[30]  = Flower( BTEX("web") ); -- Web
		[31]  = DataAdapter( 0x3, -- Tall grass
		          Flower( BTEX("deadbush") ), -- dead bush like
		          BiomeFlower( 0, BTEX("tallgrass") ), -- tall grass
		          BiomeFlower( 0, BTEX("fern") ), 0 ); -- fern
		[32]  = Flower( BTEX("deadbush") ), -- Dead bush
		-- Piston needs special code: the trick is that the side tiles need to rotate to face the front piece TODO
		[33]  = OpaqueBlock( BTEX("piston_side"), BTEX("piston_bottom"), BTEX("piston_top") ); -- Piston up
		-- [34] -- Piston Extension
		[35]  = DataAdapter( 0xf, -- Wool
		          OpaqueBlock( BTEX("cloth_0") ), -- white
		          OpaqueBlock( BTEX("cloth_1") ), -- orange
		          OpaqueBlock( BTEX("cloth_2") ), -- magenta
		          OpaqueBlock( BTEX("cloth_3") ), -- light blue
		          OpaqueBlock( BTEX("cloth_4") ), -- yellow
		          OpaqueBlock( BTEX("cloth_5") ), -- lime
		          OpaqueBlock( BTEX("cloth_6") ), -- pink
		          OpaqueBlock( BTEX("cloth_7") ), -- grey
		          OpaqueBlock( BTEX("cloth_8") ), -- light grey	
		          OpaqueBlock( BTEX("cloth_9") ), -- cyan
		          OpaqueBlock( BTEX("cloth_10") ), -- purple
		          OpaqueBlock( BTEX("cloth_11") ), -- blue
		          OpaqueBlock( BTEX("cloth_12") ), -- brown
		          OpaqueBlock( BTEX("cloth_13") ), -- green
		          OpaqueBlock( BTEX("cloth_14") ), -- red
		          OpaqueBlock( BTEX("cloth_15") ) ); -- black
		[37]  = Flower( BTEX("flower") ); -- Yellow flower
		[38]  = Flower( BTEX("rose") ); -- Red rose
		[39]  = Flower( BTEX("mushroom_brown") ); -- Brown mushroom
		[40]  = Flower( BTEX("mushroom_red") ); -- Red mushroom
		[41]  = OpaqueBlock( BTEX("blockGold") ); -- Gold block
		[42]  = OpaqueBlock( BTEX("blockIron") ); -- Iron block
		[43]  = DataAdapter( 0x7, -- Double Slab
		          DataAdapter(0x8, -- Stone
					OpaqueBlock( BTEX("stoneslab_side"), BTEX("stoneslab_top") ), -- Stone double slab
					OpaqueBlock( BTEX("stoneslab_top") ) ), -- Smooth stone slab
		          DataAdapter(0x8, -- Sandstone
					OpaqueBlock( BTEX("sandstone_side"), BTEX("sandstone_bottom"), BTEX("sandstone_top") ), -- Sandstone double slab
					OpaqueBlock( BTEX("sandstone_top") ) ), -- Smooth sandstone slab
		          OpaqueBlock( BTEX("wood") ), -- Wooden planks
		          OpaqueBlock( BTEX("stonebrick") ), -- Cobblestone
		          OpaqueBlock( BTEX("brick") ), -- Brick
		          OpaqueBlock( BTEX("stonebricksmooth") ), -- Stone brick
				  OpaqueBlock( BTEX("netherBrick") ), -- Nether brick
		          0, 0 );
		[44]  = DataAdapter( 0x8, -- Slabs
		          DataAdapter( 0x7, -- Normal Slabs
		            Slab( -8, 0, BTEX("stoneslab_side"), BTEX("stoneslab_top") ), -- Stone slab
		            Slab( -8, 0, BTEX("sandstone_side"), BTEX("sandstone_bottom"), BTEX("sandstone_top") ), -- Sandstone slab
		            Slab( -8, 0, BTEX("wood") ), -- Wooden plank slab
		            Slab( -8, 0, BTEX("stonebrick") ), -- Cobblestone slab
		            Slab( -8, 0, BTEX("brick") ), -- Brick slab
		            Slab( -8, 0, BTEX("stonebricksmooth") ), -- Stone brick
					Slab( -8, 0, BTEX("netherBrick") ), -- Nether brick
		            0, 0 ),
		          DataAdapter( 0x7, -- Upside-down slabs
		            Slab( 0, -8, BTEX("stoneslab_side"), BTEX("stoneslab_top") ), -- Stone slab
		            Slab( 0, -8, BTEX("sandstone_side"), BTEX("sandstone_bottom"), BTEX("sandstone_top") ), -- Sandstone slab
		            Slab( 0, -8, BTEX("wood") ), -- Wooden plank slab
		            Slab( 0, -8, BTEX("stonebrick") ), -- Cobblestone slab
		            Slab( 0, -8, BTEX("brick") ), -- Brick slab
		            Slab( 0, -8, BTEX("stonebricksmooth") ), -- Stone brick
					Slab( 0, -8, BTEX("netherBrick") ), -- Nether brick
		            0, 0 ) );
		[45]  = OpaqueBlock( BTEX("brick") ); -- Brick
		[46]  = OpaqueBlock( BTEX("tnt_side"), BTEX("tnt_bottom"), BTEX("tnt_top") ); -- TNT
		[47]  = OpaqueBlock( BTEX("bookshelf"), BTEX("wood") ); -- Bookshelf
		[48]  = OpaqueBlock( BTEX("stoneMoss") ); -- Mossy cobblestone
		[49]  = OpaqueBlock( BTEX("obsidian") ); -- Obsidian
		[50]  = Torch( BTEX("torch","clamp") ); -- Torch
		-- [51] -- Fire
		[52]  = HollowOpaqueBlock( BTEX("mobSpawner") ); -- Monster spawner
		[53]  = Stairs( BTEX("wood") ); -- Wooden stairs
		[54]  = CompactedBlock( -1, -1, -1, -1, 0, -2, chestSide, chestTop ); -- Chest
		-- [55] -- Redstone Wire
		[56]  = OpaqueBlock( BTEX("oreDiamond") ); -- Diamond ore
		[57]  = OpaqueBlock( BTEX("blockDiamond") ); -- Diamond block
		[58]  = OpaqueBlock( BTEX("workbench_front"), BTEX("workbench_side"), BTEX("wood"), BTEX("workbench_top") ); -- Crafting table
		[59]  = DataAdapter( 0x7, -- Wheat Crops
		          Cactus( -4, BTEX("crops_0"), 0 ), -- growthstate 0
		          Cactus( -4, BTEX("crops_1"), 0 ), -- growthstate 1
		          Cactus( -4, BTEX("crops_2"), 0 ), -- growthstate 2
		          Cactus( -4, BTEX("crops_3"), 0 ), -- growthstate 3
		          Cactus( -4, BTEX("crops_4"), 0 ), -- growthstate 4
		          Cactus( -4, BTEX("crops_5"), 0 ), -- growthstate 5
		          Cactus( -4, BTEX("crops_6"), 0 ), -- growthstate 6
		          Cactus( -4, BTEX("crops_7"), 0 ) ); -- growthstate 7 
		[60]  = DataAdapter( 0x1, -- Farmland
		          CompactedBlock( 0, 0, 0, 0, 0, -1, BTEX("dirt"), BTEX("dirt"), BTEX("farmland_dry") ), -- not hydrated
		          CompactedBlock( 0, 0, 0, 0, 0, -1, BTEX("dirt"), BTEX("dirt"), BTEX("farmland_wet") ) ); -- hydrated
		[61]  = FacingAdapter( FurnaceBody, OpaqueBlock( BTEX("furnace_front") ) ); -- Furnace
		[62]  = FacingAdapter( FurnaceBody, OpaqueBlock( BTEX("furnace_front_lit") ) ); -- Burning furnace
		-- [63] -- Sign Post
		[64]  = Door( BTEX("doorWood_lower"), BTEX("doorWood_upper") ); -- Wooden door
		[65]  = FacingAdapter( false, Cactus( -15, BTEX("ladder") ) ); -- Ladder
		[66]  = Rail( BTEX("rail"), BTEX("rail_turn") ); -- Rails
		[67]  = Stairs( BTEX("stonebrick") ); -- Cobblestone stairs
		[68]  = DataAdapter( 0x7, -- Wall Sign
		          SetTexScale( CompactedBlock( 0, 0, -14, 0, -4, -4, BTEX("wood") ), 2, 2 ), -- north
		          0, 0,
		          SetTexScale( CompactedBlock( 0, 0, 0, -14, -4, -4, BTEX("wood") ), 2, 2 ), -- south
		          SetTexScale( CompactedBlock( -14, 0, 0, 0, -4, -4, BTEX("wood") ), 2, 2 ), -- west
		          SetTexScale( CompactedBlock( 0, -14, 0, 0, -4, -4, BTEX("wood") ), 2, 2 ), -- east
		          0, 0 );
		-- [69] -- Lever
		[70]  = CompactedBlock( -1, -1, -1, -1, 0, -15, BTEX("stone") ); -- Stone pressure plate
		[71]  = Door( BTEX("doorIron_lower"), BTEX("doorIron_upper") ); -- Iron door
		[72]  = CompactedBlock( -1, -1, -1, -1, 0, -15, BTEX("wood") ); -- Wooden pressure plate
		[73]  = OpaqueBlock( BTEX("oreRedstone") ); -- Redstone ore
		[74]  = OpaqueBlock( BTEX("oreRedstone") ); -- Glowing redstone ore
		[75]  = Torch( BTEX("redtorch","clamp") ); -- Redstone torch (off)
		[76]  = Torch( BTEX("redtorch_lit","clamp") ); -- Redstone torch (on)
		[77]  = DataAdapter( 0x3+0x8, -- Stone Button
		          CompactedBlock( -5, -5, -14, 0, -6, -6, BTEX("stone") ), -- off, north
		          CompactedBlock( 0, -14, -5, -5, -6, -6, BTEX("stone") ), -- off, east
		          CompactedBlock( -14, 0, -5, -5, -6, -6, BTEX("stone") ), -- off, west
		          CompactedBlock( -5, -5, 0, -14, -6, -6, BTEX("stone") ), -- off, south
		          CompactedBlock( -5, -5, -15, 0, -6, -6, BTEX("stone") ), -- on, north
		          CompactedBlock( 0, -15, -5, -5, -6, -6, BTEX("stone") ), -- on, east
		          CompactedBlock( -15, 0, -5, -5, -6, -6, BTEX("stone") ), -- on, west
		          CompactedBlock( -5, -5, 0, -15, -6, -6, BTEX("stone") ) ); -- on, south
		[78]  = DataAdapter( 0x7, -- Snow
		          Slab( -14, 0, BTEX("snow") ), -- 1 layer
		          Slab( -12, 0, BTEX("snow") ), -- 2 layer
		          Slab( -10, 0, BTEX("snow") ), -- 3 layer
		          Slab( -8, 0, BTEX("snow") ), -- 4 layer
		          Slab( -6, 0, BTEX("snow") ), -- 5 layer
		          Slab( -4, 0, BTEX("snow") ), -- 6 layer
		          Slab( -2, 0, BTEX("snow") ), -- 7 layer
		          OpaqueBlock( BTEX("snow") ) ); -- 8 layer/block
		[79]  = DelayRender( TransparentBlock( 0, BTEX("ice") ), 5 ); -- Ice
		[80]  = OpaqueBlock( BTEX("snow") ); -- Snow block
		[81]  = Cactus( -1, BTEX("cactus_side"), BTEX("cactus_bottom"), BTEX("cactus_top") ); -- Cactus
		[82]  = OpaqueBlock( BTEX("clay") ); -- Clay
		[83]  = Flower( BTEX("reeds") ); -- Sugar Cane
		[84]  = OpaqueBlock( BTEX("musicBlock"), BTEX("musicBlock"), BTEX("jukebox_top") ); -- Jukebox
		[85]  = Fence( BTEX("wood") ); -- Wooden Fence
		[86]  = RotatingAdapter( OpaqueBlock( BTEX("pumpkin_side"), BTEX("pumpkin_top") ), OpaqueBlock( BTEX("pumpkin_face") ) ); -- Pumpkin
		[87]  = OpaqueBlock( BTEX("hellrock") ); -- Netherrack
		[88]  = OpaqueBlock( BTEX("hellsand") ); -- Soul sand
		[89]  = OpaqueBlock( BTEX("lightgem") ); -- Glowstone block
		[90]  = DelayRender( Portal( BTEX_Water("portal") ), 10 ); -- Portal
		[91]  = RotatingAdapter( OpaqueBlock( BTEX("pumpkin_side"), BTEX("pumpkin_top") ), OpaqueBlock( BTEX("pumpkin_jack") ) ); -- Jack-o-lantern
		[92]  = DataAdapter( 0x7, -- Cake
		          CompactedBlock( -1, -1, -1, -1, 0, -8, BTEX("cake_side"), BTEX("cake_bottom"), BTEX("cake_top") ), -- full cake
		          CompactedBlock( -3, -1, -1, -1, 0, -8, BTEX("cake_inner"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_bottom"), BTEX("cake_top") ), -- eaten 1
		          CompactedBlock( -5, -1, -1, -1, 0, -8, BTEX("cake_inner"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_bottom"), BTEX("cake_top") ), -- eaten 2
		          CompactedBlock( -7, -1, -1, -1, 0, -8, BTEX("cake_inner"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_bottom"), BTEX("cake_top") ), -- eaten 3
		          CompactedBlock( -9, -1, -1, -1, 0, -8, BTEX("cake_inner"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_bottom"), BTEX("cake_top") ), -- eaten 4
		          CompactedBlock( -11, -1, -1, -1, 0, -8, BTEX("cake_inner"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_side"), BTEX("cake_bottom"), BTEX("cake_top") ), 0, 0 ); -- eaten 5
		-- [93] -- Redstone repeater (off)
		-- [94] -- Redstone repeater (on)
		[95]  = CompactedBlock( -1, -1, -1, -1, 0, -2, chestSide, chestTop ); -- Locked Chest
		[96]  = DataAdapter( 0x8, -- Trapdoor
		          DataAdapter( 0x4,
		            Slab( -13, 0, BTEX("trapdoor") ), -- bottom
		            DataAdapter( 0x3,
		              CompactedBlock( 0, 0, -13, 0, 0, 0, BTEX("trapdoor") ), -- north
		              CompactedBlock( 0, 0, 0, -13, 0, 0, BTEX("trapdoor") ), -- south
		              CompactedBlock( -13, 0, 0, 0, 0, 0, BTEX("trapdoor") ), -- west
		              CompactedBlock( 0, -13, 0, 0, 0, 0, BTEX("trapdoor") ) ) ), -- east
		          DataAdapter( 0x4,
		            Slab( 0, -13, BTEX("trapdoor") ), -- top
		            DataAdapter( 0x3,
		              CompactedBlock( 0, 0, -13, 0, 0, 0, BTEX("trapdoor") ), -- north
		              CompactedBlock( 0, 0, 0, -13, 0, 0, BTEX("trapdoor") ), -- south
		              CompactedBlock( -13, 0, 0, 0, 0, 0, BTEX("trapdoor") ), -- west
		              CompactedBlock( 0, -13, 0, 0, 0, 0, BTEX("trapdoor") ) ) ) ); -- east
		[97]  = DataAdapter( 0x3, -- Hidden Silverfish
		          OpaqueBlock( BTEX("stone") ), -- stone
		          OpaqueBlock( BTEX("stonebrick") ), -- cobblestone
		          OpaqueBlock( BTEX("stonebricksmooth") ), 0 ); -- stone brick
		[98]  = DataAdapter( 0x3, -- Stone brick block
		          OpaqueBlock( BTEX("stonebricksmooth") ), -- regular
		          OpaqueBlock( BTEX("stonebricksmooth_mossy") ), -- mossy
		          OpaqueBlock( BTEX("stonebricksmooth_cracked") ), -- cracked
		          OpaqueBlock( BTEX("stonebricksmooth_carved") ) ); -- chisled
		[99] = DataAdapter( 0xf, -- Brown Mushroom
		          OpaqueBlock( BTEX("mushroom_inside") ), -- pores
		          OpaqueBlock( BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- west + north
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- north
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- east + north
		          OpaqueBlock( BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- west
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- top
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- east
		          OpaqueBlock( BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- west + south
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- south
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown"), BTEX("mushroom_inside"), BTEX("mushroom_skin_brown") ), -- east + south
		          OpaqueBlock( BTEX("mushroom_skin_stem"), BTEX("mushroom_inside") ), -- stem
		          0, 0, 0, 0, 0 );
		[100] = DataAdapter( 0xf, -- Red Mushroom
		          OpaqueBlock( BTEX("mushroom_inside") ), -- pores
		          OpaqueBlock( BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- west + north
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- north
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- east + north
		          OpaqueBlock( BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- west
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- top
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- east
		          OpaqueBlock( BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- west + south
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- south
		          OpaqueBlock( BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red"), BTEX("mushroom_inside"), BTEX("mushroom_skin_red") ), -- east + south
		          OpaqueBlock( BTEX("mushroom_skin_stem"), BTEX("mushroom_inside") ), -- stem
		          0, 0, 0, 0, 0 );
		-- [101] -- Iron Bars TODO
		[102] = HollowOpaqueBlock( BTEX("glass") ); -- Glass Panes (for now, a solid glass block - TODO)
		[103] = OpaqueBlock( BTEX("melon_side"), BTEX("melon_top") ); -- Melon
		-- [104] -- Pumpkin Stem
		-- [105] -- Melon Stem
		[106] = FaceBitAdapter( BiomeCactus( 1, {-15}, BTEX("vine"), BTEX("vine"), 0 ) ); -- Vines
		[107] = DataAdapter( 0x4, -- Fence Gate
		          DataAdapter( 0x3, -- Fence Gate CLOSED
		            MultiBlockInBlock( -- Facing South
		              { -7,  -7,  0,  -14,  -5,   0, -- End Bars
		                -7,  -7,  -14,  0,  -5,   0, 
		                -7,  -7,  -2,  -2, -12,  -1, -- Gate Parts
		                -7,  -7,  -6,  -6,  -9,  -4,
		                -7,  -7,  -2,  -2,  -6,  -7 },
		              BTEX("wood") ),
		            MultiBlockInBlock( -- Facing West
		              {-14,   0,  -7,  -7,  -5,   0, -- End Bars
		                 0, -14,  -7,  -7,  -5,   0, 
		                -2,  -2,  -7,  -7, -12,  -1, -- Gate Parts
		                -6,  -6,  -7,  -7,  -9,  -4,
		                -2,  -2,  -7,  -7,  -6,  -7 },
		              BTEX("wood") ),
		            MultiBlockInBlock( -- Facing North
		              { -7,  -7,   0, -14,  -5,   0, -- End Bars
		                -7,  -7, -14,   0,  -5,   0, 
		                -7,  -7,  -2,  -2,  -12, -1, -- Gate Parts
		                -7,  -7,  -6,  -6,  -9,  -4,
		                -7,  -7,  -2,  -2,  -6,  -7 },
		              BTEX("wood") ),
		            MultiBlockInBlock( -- Facing East
		              {-14,   0,  -7,  -7,  -5,   0, -- End Bars
		                 0, -14,  -7,  -7,  -5,   0, 
		                -2,  -2,  -7,  -7, -12,  -1, -- Gate Parts
		                -6,  -6,  -7,  -7,  -9,  -4,
		                -2,  -2,  -7,  -7,  -6,  -7 },
		              BTEX("wood") ) ),
		          DataAdapter( 0x3, -- Fence Gate OPEN
		            MultiBlockInBlock( -- Facing South
		              { -7,  -7,   0,  -14,  -5,  0, -- End Bars
		                -7,  -7, -14,   0,  -5,   0, 
		                -9,  -1,   0, -14, -12,  -1, -- Gate Right
		               -13,  -1,   0, -14,  -9,  -4,
		                -9,  -1,   0, -14,  -6,  -7,
		                -9,  -1, -14,   0, -12,  -1, -- Gate Left
		               -13,  -1, -14,   0,  -9,  -4,
		                -9,  -1, -14,   0,  -6,  -7 },
		              BTEX("wood") ),
		            MultiBlockInBlock( -- Facing West
		              {-14,   0,  -7,  -7,  -5,   0, -- End Bars
		                 0, -14,  -7,  -7,  -5,   0, 
		               -14,   0,  -1,  -9, -12,  -1, -- Gate Left
		               -14,   0,  -1, -13,  -9,  -4,
		               -14,   0,  -1,  -9,  -6,  -7,
		                 0, -14,  -1,  -9, -12,  -1, -- Gate Right
		                 0, -14,  -1, -13,  -9,  -4,
		                 0, -14,  -1,  -9,  -6,  -7 },
		              BTEX("wood") ),
		            MultiBlockInBlock( -- Facing North
		              { -7,  -7,   0, -14,  -5,   0, -- End Bars
		                -7,  -7, -14,   0,  -5,   0, 
		                -1,  -9,   0, -14, -12,  -1, -- Gate Left
		                -1, -13,   0, -14,  -9,  -4,
		                -1,  -9,   0, -14,  -6,  -7,
		                -1,  -9, -14,   0, -12,  -1, -- Gate Right
		                -1, -13, -14,   0,  -9,  -4,
		                -1,  -9, -14,   0,  -6,  -7 },
		              BTEX("wood") ),
		            MultiBlockInBlock( -- Facing East
		              {-14,   0,  -7,  -7,  -5,   0, -- End Bars
		                 0, -14,  -7,  -7,  -5,   0, 
		               -14,   0,  -9,  -1, -12,  -1, -- Gate Right
		               -14,   0, -13,  -1,  -9,  -4,
		               -14,   0,  -9,  -1,  -6,  -7,
		                 0, -14,  -9,  -1, -12,  -1, -- Gate Left
		                 0, -14, -13,  -1,  -9,  -4,
		                 0, -14,  -9,  -1,  -6,  -7 },
		              BTEX("wood") ) ) );
		[108] = Stairs( BTEX("brick") ); -- Brick Stairs
		[109] = Stairs( BTEX("stonebricksmooth") ); -- Stone Brick Stairs
		[110] = OpaqueBlock( BTEX("mycel_side"), BTEX("dirt"), BTEX("mycel_top") ); -- Mycelium
		[111] = BiomeCactus( 0, {-15}, 0, 0, BTEX("waterlily") ); -- Lily pad
		[112] = OpaqueBlock( BTEX("netherBrick") ); -- Nether Brick
		[113] = Fence( BTEX("netherBrick") ); -- Nether Brick Fence
		[114] = Stairs( BTEX("netherBrick") ); -- Nether Brick Stairs
		[115] = DataAdapter( 0x3, -- Nether Wart
		          Cactus( -4, BTEX("netherStalk_0"), 0 ), -- growthstate 0
		          Cactus( -4, BTEX("netherStalk_1"), 0 ), 1, -- growthstate 1
		          Cactus( -4, BTEX("netherStalk_2"), 0 ) ); -- growthstate 2
		[116] = Slab( -4, 0, BTEX("enchantment_side"), BTEX("enchantment_bottom"), BTEX("enchantment_top") ); -- Enchantment Table
		-- [117] -- Brewing Stand
		-- [118] -- Cauldron
		-- [119] -- End Portal
		-- [120] -- End Portal Frame
		[121] = OpaqueBlock( BTEX("whiteStone") ); -- End Stone
		[122] = MultiBlockInBlock( -- Dragon Egg
		          { -6, -6, -6, -6, -15,   0,
		            -5, -5, -5, -5, -14,  -1,
		            -4, -4, -4, -4, -13,  -2,
		            -3, -3, -3, -3, -11,  -3,
		            -2, -2, -2, -2,  -8,  -5,
		            -1, -1, -1, -1,  -3,  -8,
		            -3, -3, -3, -3,  -1, -13,
		            -6, -6, -6, -6,   0, -15}, -- Egg
		          BTEX("dragonEgg") );
		[123] = OpaqueBlock( BTEX("redstoneLight") ); -- Redstone Lamp (off)
		[124] = OpaqueBlock( BTEX("redstoneLight_lit") ); -- Redstone Lamp (on)
		[125] = DataAdapter( 0x3, -- Wooden Double-Slab
		          OpaqueBlock( BTEX("wood") ), -- oak
		          OpaqueBlock( BTEX("wood_spruce") ), -- spruce
		          OpaqueBlock( BTEX("wood_birch") ), -- birch
		          OpaqueBlock( BTEX("wood_jungle") ) ); -- jungle
		[126] = DataAdapter( 0x3+0x8, -- Wooden Slab
		          Slab( -8, 0, BTEX("wood") ), -- bottom, oak
		          Slab( -8, 0, BTEX("wood_spruce") ), -- bottom, spruce
		          Slab( -8, 0, BTEX("wood_birch") ), -- bottom, birch
		          Slab( -8, 0, BTEX("wood_jungle") ), -- bottom, jungle
		          Slab( 0, -8, BTEX("wood") ), -- top, oak
		          Slab( 0, -8, BTEX("wood_spruce") ), -- top, spruce
		          Slab( 0, -8, BTEX("wood_birch") ), -- top, birch
		          Slab( 0, -8, BTEX("wood_jungle") ) ); -- top, jungle
		-- [127] -- Cocoa Plant
		[128] = Stairs( BTEX("sandstone_side"), BTEX("sandstone_bottom"), BTEX("sandstone_top") ); -- Sandstone Stairs
		[129] = OpaqueBlock( BTEX("oreEmerald") ); -- Emerald Ore
		-- [130] -- Ender Chest
		-- [131] -- Tripwire Hook
		-- [132] -- Tripwire
		[133] = OpaqueBlock( BTEX("blockEmerald") ); -- Block of Emerald
		[134] = Stairs( BTEX("wood_spruce") ); -- Spruce Wood Stairs
		[135] = Stairs( BTEX("wood_birch") ); -- Birch Wood Stairs
		[136] = Stairs( BTEX("wood_jungle") ); -- Jungle Wood Stairs
		[137] = OpaqueBlock( BTEX("commandBlock") ); -- Command Block
		-- [138] -- Beacon Block
		-- [139] -- Cobblestone Wall
		-- [140] -- Flower Pot
		[141] = DataAdapter( 0x7, -- Carrots
		          Cactus( -4, BTEX("carrots_0"), 0 ), Cactus( -4, BTEX("carrots_0"), 0 ), -- growthstate 0
		          Cactus( -4, BTEX("carrots_1"), 0 ), Cactus( -4, BTEX("carrots_1"), 0 ), -- growthstate 1
		          Cactus( -4, BTEX("carrots_2"), 0 ), Cactus( -4, BTEX("carrots_2"), 0 ), Cactus( -4, BTEX("carrots_2"), 0 ), -- growthstate 2
		          Cactus( -4, BTEX("carrots_3"), 0 ) ); -- growthstate 3
		[142] = DataAdapter( 0x7, -- Potatoes
		          Cactus( -4, BTEX("potatoes_0"), 0 ), Cactus( -4, BTEX("potatoes_0"), 0 ), -- growthstate 0
		          Cactus( -4, BTEX("potatoes_1"), 0 ), Cactus( -4, BTEX("potatoes_1"), 0 ), -- growthstate 1
		          Cactus( -4, BTEX("potatoes_2"), 0 ), Cactus( -4, BTEX("potatoes_2"), 0 ), Cactus( -4, BTEX("potatoes_2"), 0 ), -- growthstate 2
		          Cactus( -4, BTEX("potatoes_3"), 0 ) ); -- growthstate 3
		[143] = DataAdapter( 0x3+0x8, -- Wooden Button
		          CompactedBlock( -5, -5, -14, 0, -6, -6, BTEX("wood") ), -- off, north
		          CompactedBlock( 0, -14, -5, -5, -6, -6, BTEX("wood") ), -- off, east
		          CompactedBlock( -14, 0, -5, -5, -6, -6, BTEX("wood") ), -- off, west
		          CompactedBlock( -5, -5, 0, -14, -6, -6, BTEX("wood") ), -- off, south
		          CompactedBlock( -5, -5, -15, 0, -6, -6, BTEX("wood") ), -- on, north
		          CompactedBlock( 0, -15, -5, -5, -6, -6, BTEX("wood") ), -- on, east
		          CompactedBlock( -15, 0, -5, -5, -6, -6, BTEX("wood") ), -- on, west
		          CompactedBlock( -5, -5, 0, -15, -6, -6, BTEX("wood") ) ); -- on, south
		-- [144] -- Head Block
		-- [145] -- Anvil
		-- [146] -- Trapped Chest
		[147] = CompactedBlock( -1, -1, -1, -1, 0, -15, BTEX("blockGold") ); -- Weighted Pressure Plate (Light)
		[148] = CompactedBlock( -1, -1, -1, -1, 0, -15, BTEX("blockIron") ); -- Weighted Pressure Plate (Heavy)
		-- [149] -- Redstone Comparator (inactive)
		-- [150] -- Redstone Comparator (active)
		[151] = CompactedBlock( 0, 0, 0, 0, 0, -10, BTEX("daylightDetector_side"), BTEX("daylightDetector_side"), BTEX("daylightDetector_top") ); -- Daylight Sensor
		[152] = OpaqueBlock( BTEX("blockRedstone") ); -- Block of Redstone
		[153] = OpaqueBlock( BTEX("netherquartz") ); -- Nether Quartz (Ore)
		-- [154] -- Hopper
		-- [155] -- Block of Quartz
		-- [156] -- Quartz Stairs
		-- [157] -- Activator Rail
		-- [158] -- Dropper
		-- [170] -- Hay Bale
		-- [171] -- Carpet
		-- [172] -- Hardened Clay
	};

	for id, blk in pairs( MinecraftBlocks ) do
		blocks:setGeometry( id, blk[1] );
		blocks:setSolidity( id, blk[2] );
	end

	return blocks;
end
